/***Copyright (c) 1997-2017  Alexandros V. Gerbessiotis.
 **** All rights reserved.
 **** ****** Bulk-Synchronous Parallel Sorting
 **** ****** An implementation of the
 **** ****** Gerbessiotis-Siniolakis deterministic sorting 
 **** ****** Algorithm. Single round implementation only.
 ****Permission to use, copy, modify, and distribute this software,
 ****and to incorporate it, in whole or in part, into
 ****other software,is hereby granted without fee, provided that
 ****  (1) the above copyright notice and this permission
 ****      notice appear in all copies of the source code, and
 ****      the above copyright notice appear in clearly visible
 ****      form on all supporting documentation and distribution
 ****      media, and
 ****     (2) any redistribution of the software, in original
 ****         or modified form, be without fee and subject to
 ****         these same conditions.
 ****   No guarantee, whatsoever, is offered that the code works as
 ****   advertised or is absent of any, even damaging,
 ****   side-effects. Use this code at your own personal risk.
 ****/

#include <stdlib.h>
#include "avgai.h"
#include "prdx.h"
#include "psrtai.h"
#include "seqo.h"
#include <math.h>

#define INITTIME(Y)   Y=AITIME()
#define PRINTTIME(X,Y) AIBARRIER();Y=AITIME();\
                  if (0==pid) \
                     {fprintf(stdout,#X"%10.4f\n",(Y-t0));fflush(stdout);}

int bspbtndat (idata *,idata *,int)                         ;
void aiscan   (int   *,int   *,int)                         ;


 /**** gsd       :Gerbessiotis-Siniolakis Deterministic sorting
  **** keys      : array of idata to sort, idata is int or double
  ****    n      : # of keys in local processor
  **** total     : total number of keys;may not be n*nprocs.If total==0
  ****             then total is set to n*nprocs; 
  **** result    : Output array: Algorithm does not sort in place
  **** nbytes    : byte size of idata nbytes=sizeof(idata)
  **** localsort : Sequential Sort function to sort local idata
  **** generic    : Generic Sort function
  **** compare   : compare function to compare two idata keys
  **** percent   : of actual sample size; allows control of sample
  ****             size to vary from theoretical value.
  **** Returns an integer that is the number of local keys after sort.
  ****/
int  gsd (idata  *keys, int n, int total, idata **result,
	  int nbytes, void (*localsort)(idata *,int), 
          void (*generic)(void*,int,int,int (*)(const void*,const void *)),
          int  (*comparerec)(const void  *,const void  *), 
          int  (*bcompare)(idata*,record*,int,int), 
          int percent)

{
  register int  pid;    /* processor id */
  register int  nprocs; /* number of processors */
  register double u;    /* auxiliary variable */
  register int  ns;     /* sample size */
  idata          *sample;          /* key sample array */
  idata          *ssample;          /* key sample array */
  idata          *split;           /* splitter array */
  int           *bucket;          /* local bucket size */
  int           *obucket;         /* auxiliary local */
  int           *ppf_bucket;      /* ParallelPrefix'ed bucket size */
  int           *lbucket;         /* auxiliary local */
  register int  offset,index;     /* aux variables in random sampling*/
  register int  proc_size;        /* processor bucket size */
  register int  size,intsize;     /* size of idata and int */
  register int  i,j  ;            /* index variables */
  register int  left,right,middle;/* binary search variables */
  idata          *temp_result;
#ifdef XTIME
  double t0,t1,t2;
#endif

  /**** PART 0: Some early checks*/
  pid=AIPID(); nprocs=AINPROCS();
  size=nbytes;
#ifdef SANITY_CHECK
  if ((n <= 0) || (total <0))  {
     AIABORT("gsd:: Problem size negative or zero\n");
  }
#endif
  if (nprocs <= 1) {
    temp_result = (idata  *) malloc(n*size);
#ifdef SANITY_CHECK
    check_if_null((void *) temp_result,"gsd","temp_result");
#endif
    memcpy(temp_result,keys,size*n);
    (*localsort) ((idata *)temp_result,n);
    *result=(idata *)temp_result;
    return(n);	
  }

  /**** PART 1: Initializations */

#ifdef XTIME
  INITTIME(t0);
#endif
#ifdef PARALLELSAMPLESORT
  if (total==0) total=nprocs*n;
  u=(percent/100.0)*nprocs*log(log((double)total));
  u=(percent/100.0)*nprocs*log((double)total);
  u=ceil(u+1); /* auxiliary results */
  ns= (((int) u)*nprocs);/* total uample size */
  intsize=sizeof(int);   /* sizeof(int)   */
#else
  if (total==0) total=nprocs*n;
  u=(percent/100.0)*nprocs;
  ns= (((int) u)*nprocs);/* total sample size */
  intsize=sizeof(int);   /* sizeof(int)   */
#endif

  /**** PART 2: Random number initialization */

  /* It's the deterministic algorithm; What did you expect ? */ 

  /**** PART 3: Space allocation and checking */

  sample = (idata *) malloc(ns*size);     /* Sample idata storage */
  ssample = (idata *) malloc(ns*size);     /* Sample idata storage */
  split  = (idata *) malloc(nprocs*size); /* splitter storage */
  bucket = (int *)  malloc(nprocs*intsize); /* bucket size */
  obucket = (int *)  malloc(nprocs*intsize); /* bucket size */
  ppf_bucket = (int *) malloc(nprocs*intsize); /* ppf result temp */
  lbucket = (int *) malloc(nprocs*intsize); /* auxiliary variable */
#ifdef SANITY_CHECK
  check_if_null((void *) sample,    "gsd","sample    ");
  check_if_null((void *) ssample,   "gsd","ssample   ");
  check_if_null((void *) split ,    "gsd","split     " );
  check_if_null((void *) bucket,    "gsd","bucket    ");
  check_if_null((void *) obucket,   "gsd","obucket   ");
  check_if_null((void *) ppf_bucket,"gsd","ppf_bucket");
  check_if_null((void *) lbucket   ,"gsd","lbucket   ");
#endif

  /**** PART 4: Registration area */
  /* Delayed registration; intime registration before communication*/

  /**** PART 5: Regular OVERsampling*/

  /* Example: Say n= 11 localsamplesize=ns/nprocs=3
   o  o  o  o  o  o  o  o  o  o  o   
            ^           ^        ^
   sample   1           2        3=maximum
   CEIL(...) =4 FLOOR(...) =3 i.e. interspace is
              3             2  respectively
  */
#ifdef XTIME
  PRINTTIME("Check 1(Initial  )",t1);
#endif

  (*localsort)((idata *)keys,n); /****Sort first; Sample later ****/

#ifdef XTIME
  PRINTTIME("Check 2(LocalSort)",t1);
#endif

  index=LEFT(n,(ns/nprocs));/* Start sampling;Store sample in sample*/
  offset=CEIL(n,(ns/nprocs));/* equidistant sample keys are picked  */ 
  for(i=0;i<index;i++) { /* ceiling(ns/nprocs) for the first keys*/
    sample[i] = keys[(i+1)*offset-1];
  }
  offset=FLOOR(n,(ns/nprocs));/* floor(ns/nprocs) for remainder*/
  for(i=index;i<((ns/nprocs)-1);i++) {
  sample[i] = keys[(i+1)*offset-1];
  }

                         /* Append the maximum to the sample */
  sample[ns/nprocs-1] = keys[(ns/nprocs)*offset-1];
  /*or  sample[i          ] = keys[(i+1      )*offset-1]; */

  offset=(ns/nprocs)*pid*size;
  bucket[0]=ns/nprocs; /* Sample size per processor */

  /**** PART 6: Sample sorting */
#ifndef PARALLELSAMPLESORT
  AIREGISTER(sample,ns*size);
  AIINIT();AICOMMIT();
  /* collect sample at processor 0 */
  if (pid != 0 )
    AIPUT(0,sample,sample,offset,bucket[0]*size);
  AISYNC();AIBARRIER();

  /* sort sample  and select p-1 splitters */
  if (pid == 0) {
    (*localsort)((idata*)sample,ns);
    for(i=1;i<nprocs;i++)
       split[i-1]=sample[(ns*i/nprocs)-1];
  }
  AIREGISTER(split,nprocs*size);
  AIINIT();AICOMMIT();
  if (pid==0){
    for(i=1;i<nprocs;i++){/*Last local sample becomes a splitter*/
      AIHPPUT(i,split,split,0,(nprocs-1)*size);
    }
  }
  AISYNC();AIBARRIER();
#else
  bucket[0]=
    bspbtndat(sample,ssample,bucket[0]);

    /* Done with sample sorting; Locate splitters*/
    /* Locate splitters in resulting merged sequence */
  AIREGISTER(split,nprocs*size);
  AIINIT();AICOMMIT();
  for(i=0;i<nprocs;i++){/*Last local sample becomes a splitter*/
    AIHPPUT(i,&ssample[(bucket[0]-1)],split,pid*size,size);
  }
  AISYNC();AIBARRIER();
#endif

  /**** PART 7: Splitters and Buckets preparation */
  for(i=0;i<nprocs;i++) {	
     bucket[i]=0;
  }

  /**** PART 8: Binary search of keys into the splitters */
  /**** It is more efficient to do it the other way around!!*/
  /**** # of splitters = nprocs-1; #of buckets = nprocs  */
  for(i=nprocs-2,bucket[nprocs-1]=n;i>=0;i--){
      bucket[i]=searchbdat((idata *)&split[i],(idata *)keys,n);
      bucket[i+1]-=bucket[i];
  }
     
  /**** PART 9: Independent prefix operations on bucket sizes */
  AI1REGISTER(ppf_bucket,nprocs*intsize);
  AI1INIT(ppf_bucket);AICOMMIT();
  aiscan((int  *)bucket,(int  *)ppf_bucket,nprocs);
/*
  ai2prefix(operadd,nprocs,(char *)bucket,(char *)ppf_bucket,intsize);
  if (pid==(2)){
   for(k=0;k<nprocs;k++) printf("%d %d %d\n",k,bucket[k],ppf_bucket[k]);
  }
*/

  /**** PART 10: Rearrange keys according to bucket destination */
  /* Send bucket size to corresponding processors */
  AIREGISTER(lbucket,nprocs*intsize);
  AIINIT();AICOMMIT();
  for(i=0;i<nprocs;i++) {
     AIHPPUT(i,&ppf_bucket[i],lbucket,pid*intsize,intsize);
  }
  AISYNC();AIBARRIER();

#ifdef XTIME
  PRINTTIME("Check 4",t1);
#endif
  /* Allocate memory */
  proc_size = lbucket[nprocs-1];
  temp_result = (idata *) malloc(proc_size*size);
#ifdef SANITY_CHECK
  check_if_null((void *)temp_result,"gsd","temp_result");
#endif
   AIREGISTER(temp_result,proc_size*size);
   AIINIT();AICOMMIT();

   /**** PART 11: Route keys to destination processor */
#ifdef ROUTEONE
   for(i=0,j=0;i<nprocs;i++) {
        middle=bucket[i]*size;
        right=ppf_bucket[i]*size;
        left=right-middle;
        AIHPPUT(i,&keys[j],temp_result,left,middle);
/*
        for(k=0;k<(n/256000)*400;k++) obucket[i] +=k;
*/
        j+=bucket[i];
   }
   AISYNC(); /* AIBARRIER(); */
#else
  obucket[0]=0;
  for(i=1;i<nprocs;i++) {
    obucket[i]=obucket[i-1]+bucket[i-1];
  }
  AIBARRIER();
   for(k=0,j=0;k<nprocs;k++) {
        i=(k+1)%nprocs;
        middle=bucket[i]*size;
        right=ppf_bucket[i]*size;
        left=right-middle;
        AIHPPUT(i,&keys[obucket[i]],temp_result,left,middle);
        j+=bucket[i];
   }
   AISYNC();AIBARRIER();
#endif

#ifdef XTIME
  PRINTTIME("Check 5(Routing)",t1);
#endif

   /**** PART 11a: Start merging received sequences */
#ifdef KWAYMERGE
   for(i=1;i<nprocs; i = (i<<1)) {
      for(left=0,j=0;left < nprocs;left += (i<<1)) {
         middle = MIN((left+(i<<1)),nprocs) -1;
         right=lbucket[left+i-1];
         mergedat2((idata *)&temp_result[j],
               (idata *) &temp_result[right],
               (right-j),
               (lbucket[middle]-right)
               );
         j=lbucket[middle];
      }
   }
#else
   (*localsort)((idata*)temp_result,proc_size);
/*
   sortdat_mrg((idata*)temp_result,proc_size);
*/
#endif

#ifdef XTIME
  PRINTTIME("Check 6(Merging)",t1);
#endif

   /**** PART 12: Ready to exit. Free space and Deregister*/
  AIDEREGISTER(temp_result);
  AIDEREGISTER(lbucket);
  AIDEREGISTER(ppf_bucket);
  AIDEREGISTER(split);
  free((void*)lbucket);
  free((void*)ppf_bucket); 
  free((void*)obucket);
  free((void*)bucket);
  free((void*)split);
  free((void*)sample);

#ifdef XTIME
  PRINTTIME("Check 7(Done)",t1);
#endif
  *result=temp_result;
  return(proc_size);

}
